"""
This module defines Exception classes used by Connexion to generate a proper response.
"""

import typing as t

from jsonschema.exceptions import ValidationError
from starlette.exceptions import HTTPException

from .problem import problem


class ConnexionException(Exception):
    """Base class for any exception thrown by the Connexion framework."""


class ResolverError(LookupError, ConnexionException):
    """Error raised at startup when the resolver cannot find a view function for an endpoint in
    your specification, and no ``resolver_error`` is configured."""


class InvalidSpecification(ValidationError, ConnexionException):
    """Error raised at startup when the provided specification cannot be validated."""


class MissingMiddleware(ConnexionException):
    """Error raised when you're leveraging behavior that depends on a specific middleware,
    and that middleware is not part of your middleware stack."""


# HTTP ERRORS


class ProblemException(HTTPException, ConnexionException):
    """
    This exception holds arguments that are going to be passed to the
    `connexion.problem` function to generate a proper response.
    """

    def __init__(
        self,
        *,
        status=500,
        title=None,
        detail=None,
        type=None,
        instance=None,
        headers=None,
        ext=None,
    ):
        self.status = self.status_code = status
        self.title = title
        self.detail = detail
        self.type = type
        self.instance = instance
        self.headers = headers
        self.ext = ext

    def to_problem(self):
        return problem(
            status=self.status,
            title=self.title,
            detail=self.detail,
            type=self.type,
            instance=self.instance,
            headers=self.headers,
            ext=self.ext,
        )


# CLIENT ERRORS (4XX)


class ClientProblem(ProblemException):
    """Base exception for any 4XX error. Returns 400 by default, however
    :class:`BadRequestProblem` should be preferred for 400 errors."""

    def __init__(self, status: int = 400, title: str = None, *, detail: str = None):
        super().__init__(status=status, title=title, detail=detail)


class BadRequestProblem(ClientProblem):
    """Problem class for 400 Bad Request errors."""

    def __init__(self, detail=None):
        super().__init__(status=400, title="Bad Request", detail=detail)


class ExtraParameterProblem(BadRequestProblem):
    """Problem class for 400 Bad Request errors raised when extra query or form parameters are
    detected and ``strict_validation`` is enabled."""

    def __init__(self, *, param_type: str, extra_params: t.Iterable[str]):
        detail = f"Extra {param_type} parameter(s) {','.join(extra_params)} not in spec"
        super().__init__(detail=detail)


class TypeValidationError(BadRequestProblem):
    """Problem class for 400 Bad Request errors raised when path, query or form parameters with
    an incorrect type are detected."""

    def __init__(self, schema_type: str, parameter_type: str, parameter_name: str):
        detail = f"Wrong type, expected '{schema_type}' for {parameter_type} parameter '{parameter_name}'"
        super().__init__(detail=detail)


class Unauthorized(ClientProblem):
    """Problem class for 401 Unauthorized errors."""

    description = (
        "The server could not verify that you are authorized to access"
        " the URL requested. You either supplied the wrong credentials"
        " (e.g. a bad password), or your browser doesn't understand"
        " how to supply the credentials required."
    )

    def __init__(self, detail: str = description):
        super().__init__(401, title="Unauthorized", detail=detail)


class OAuthProblem(Unauthorized):
    """Problem class for 401 Unauthorized errors raised when there is an issue with the received
    OAuth headers."""

    pass


class OAuthResponseProblem(OAuthProblem):
    """Problem class for 401 Unauthorized errors raised when improper OAuth credentials are
    retrieved from your OAuth server."""

    pass


class Forbidden(HTTPException):
    """Problem class for 403 Unauthorized errors."""

    def __init__(self, detail: t.Optional[str] = None):
        if detail is None:
            detail = (
                "You don't have the permission to access the requested"
                " resource. It is either read-protected or not readable by the"
                " server."
            )
        super().__init__(403, detail=detail)


class OAuthScopeProblem(Forbidden):
    """Problem class for 403 Unauthorized errors raised because of OAuth scope validation errors."""

    def __init__(self, token_scopes: list, required_scopes: list) -> None:
        self.required_scopes = required_scopes
        self.token_scopes = token_scopes
        detail = (
            f"Provided token does not have the required scopes. "
            f"Provided: {token_scopes}; Required: {required_scopes}"
        )
        super().__init__(detail=detail)


class UnsupportedMediaTypeProblem(ClientProblem):
    """Problem class for 415 Unsupported Media Type errors which are raised when Connexion
    receives a request with an unsupported media type header."""

    def __init__(self, detail: t.Optional[str] = None):
        super().__init__(status=415, title="Unsupported Media Type", detail=detail)


# SERVER ERRORS (5XX)


class ServerError(ProblemException):
    """Base exception for any 5XX error. Returns 500 by default, however
    :class:`InternalServerError` should be preferred for 500 errors."""

    def __init__(
        self,
        status: int = 500,
        title: t.Optional[str] = None,
        *,
        detail: t.Optional[str] = None,
    ):
        if title is None:
            title = "Internal Server Error"

        super().__init__(status=status, title=title, detail=detail)


class InternalServerError(ServerError):
    """Problem class for 500 Internal Server errors."""

    def __init__(self, detail: t.Optional[str] = None):
        if detail is None:
            detail = (
                "The server encountered an internal error and was unable to complete your "
                "request. Either the server is overloaded or there is an error in the application."
            )
        super().__init__(status=500, title="Internal Server Error", detail=detail)


class NonConformingResponse(InternalServerError):
    """Problem class for 500 Internal Server errors raised because of a returned response not
    matching the specification if response validation is enabled."""

    def __init__(self, detail: t.Optional[str] = None):
        super().__init__(detail=detail)


class NonConformingResponseBody(NonConformingResponse):
    """Problem class for 500 Internal Server errors raised because of a returned response body not
    matching the specification if response validation is enabled."""

    def __init__(self, detail: t.Optional[str] = None):
        if detail is None:
            detail = "Response body does not conform to specification"

        super().__init__(detail=detail)


class NonConformingResponseHeaders(NonConformingResponse):
    """Problem class for 500 Internal Server errors raised because of a returned response headers
    not matching the specification if response validation is enabled."""

    def __init__(self, detail: t.Optional[str] = None):
        if detail is None:
            detail = "Response headers do not conform to specification"

        super().__init__(detail=detail)


class ResolverProblem(ServerError):
    """Problem class for 501 Not Implemented errors raised when the resolver cannot find a view
    function to handle the incoming request."""

    def __init__(self, status: int = 501, *, detail: t.Optional[str] = None):
        super().__init__(status=status, title="Not Implemented", detail=detail)



from fnmatch import fnmatch

# special marker object to return empty content for any status code
# e.g. in app method do "return NoContent, 201"
NoContent = object()


class MediaTypeDict(dict):
    """
    A dictionary where keys can be either media types or media type ranges. When fetching a
    value from the dictionary, the provided key is checked against the ranges. The most specific
    key is chosen as prescribed by the OpenAPI spec, with `type/*` being preferred above
    `*/subtype`.
    """

    def __getitem__(self, item):
        # Sort keys in order of specificity
        for key in sorted(self, key=lambda k: ("*" not in k, k), reverse=True):
            if fnmatch(item, key):
                return super().__getitem__(key)
        raise super().__getitem__(item)

    def get(self, item, default=None):
        try:
            return self[item]
        except KeyError:
            return default

    def __contains__(self, item):
        try:
            self[item]
        except KeyError:
            return False
        else:
            return True



import abc
import asyncio
import functools
import json
import typing as t

from asgiref.sync import async_to_sync
from starlette.concurrency import run_in_threadpool

from connexion.context import operation, receive, scope
from connexion.decorators.parameter import (
    AsyncParameterDecorator,
    BaseParameterDecorator,
    SyncParameterDecorator,
)
from connexion.decorators.response import (
    AsyncResponseDecorator,
    BaseResponseDecorator,
    NoResponseDecorator,
    SyncResponseDecorator,
)
from connexion.frameworks.abstract import Framework
from connexion.frameworks.starlette import Starlette as StarletteFramework
from connexion.uri_parsing import AbstractURIParser
from connexion.utils import not_installed_error

try:
    from connexion.frameworks.flask import Flask as FlaskFramework
except ImportError as e:
    _flask_not_installed_error = not_installed_error(
        e, msg="Please install connexion using the 'flask' extra"
    )
    FlaskFramework = _flask_not_installed_error  # type: ignore


class BaseDecorator:
    """Base class for connexion decorators."""

    framework: t.Type[Framework]

    def __init__(
        self,
        *,
        pythonic_params: bool = False,
        uri_parser_class: AbstractURIParser = None,
        jsonifier=json,
    ) -> None:
        self.pythonic_params = pythonic_params
        self.uri_parser_class = uri_parser_class
        self.jsonifier = jsonifier

        self.arguments, self.has_kwargs = None, None

    @property
    @abc.abstractmethod
    def _parameter_decorator_cls(self) -> t.Type[BaseParameterDecorator]:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def _response_decorator_cls(self) -> t.Type[BaseResponseDecorator]:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def _sync_async_decorator(self) -> t.Callable[[t.Callable], t.Callable]:
        """Decorator to translate between sync and async functions."""
        raise NotImplementedError

    @property
    def uri_parser(self):
        uri_parser_class = self.uri_parser_class or operation.uri_parser_class
        return uri_parser_class(operation.parameters, operation.body_definition())

    def decorate(self, function: t.Callable) -> t.Callable:
        """Decorate a function with decorators based on the operation."""
        function = self._sync_async_decorator(function)

        parameter_decorator = self._parameter_decorator_cls(
            framework=self.framework,
            pythonic_params=self.pythonic_params,
        )
        function = parameter_decorator(function)

        response_decorator = self._response_decorator_cls(
            framework=self.framework,
            jsonifier=self.jsonifier,
        )
        function = response_decorator(function)

        return function

    @abc.abstractmethod
    def __call__(self, function: t.Callable) -> t.Callable:
        raise NotImplementedError


class WSGIDecorator(BaseDecorator):
    """Decorator for usage with WSGI apps. The parameter decorator works with a Flask request,
    and provides Flask datastructures to the view function. This works for any WSGI app, since
    we get the request via the connexion context provided by WSGI middleware.

    This decorator does not parse responses, but passes them directly to the WSGI App."""

    framework = FlaskFramework

    @property
    def _parameter_decorator_cls(self) -> t.Type[SyncParameterDecorator]:
        return SyncParameterDecorator

    @property
    def _response_decorator_cls(self) -> t.Type[BaseResponseDecorator]:
        return NoResponseDecorator

    @property
    def _sync_async_decorator(self) -> t.Callable[[t.Callable], t.Callable]:
        def decorator(function: t.Callable) -> t.Callable:
            @functools.wraps(function)
            def wrapper(*args, **kwargs) -> t.Callable:
                if asyncio.iscoroutinefunction(function):
                    return async_to_sync(function)(*args, **kwargs)
                else:
                    return function(*args, **kwargs)

            return wrapper

        return decorator

    def __call__(self, function: t.Callable) -> t.Callable:
        @functools.wraps(function)
        def wrapper(*args, **kwargs):
            request = self.framework.get_request(uri_parser=self.uri_parser)
            decorated_function = self.decorate(function)
            return decorated_function(request)

        return wrapper


class FlaskDecorator(WSGIDecorator):
    """Decorator for usage with Connexion or Flask apps. The parameter decorator works with a
    Flask request, and provides Flask datastructures to the view function.

    The response decorator returns Flask responses."""

    @property
    def _response_decorator_cls(self) -> t.Type[SyncResponseDecorator]:
        return SyncResponseDecorator


class ASGIDecorator(BaseDecorator):
    """Decorator for usage with ASGI apps. The parameter decorator works with a Starlette request,
    and provides Starlette datastructures to the view function. This works for any ASGI app, since
    we get the request via the connexion context provided by ASGI middleware.

    This decorator does not parse responses, but passes them directly to the ASGI App."""

    framework = StarletteFramework

    @property
    def _parameter_decorator_cls(self) -> t.Type[AsyncParameterDecorator]:
        return AsyncParameterDecorator

    @property
    def _response_decorator_cls(self) -> t.Type[BaseResponseDecorator]:
        return NoResponseDecorator

    @property
    def _sync_async_decorator(self) -> t.Callable[[t.Callable], t.Callable]:
        def decorator(function: t.Callable) -> t.Callable:
            @functools.wraps(function)
            async def wrapper(*args, **kwargs):
                if asyncio.iscoroutinefunction(function):
                    return await function(*args, **kwargs)
                else:
                    return await run_in_threadpool(function, *args, **kwargs)

            return wrapper

        return decorator

    def __call__(self, function: t.Callable) -> t.Callable:
        @functools.wraps(function)
        async def wrapper(*args, **kwargs):
            request = self.framework.get_request(
                uri_parser=self.uri_parser, scope=scope, receive=receive
            )
            decorated_function = self.decorate(function)
            return await decorated_function(request)

        return wrapper


class StarletteDecorator(ASGIDecorator):
    """Decorator for usage with Connexion or Starlette apps. The parameter decorator works with a
    Starlette request, and provides Starlette datastructures to the view function.

    The response decorator returns Starlette responses."""

    @property
    def _response_decorator_cls(self) -> t.Type[AsyncResponseDecorator]:
        return AsyncResponseDecorator



"""
This module defines a native connexion asynchronous application.
"""

import functools
import logging
import pathlib
import typing as t

from starlette.responses import Response as StarletteResponse
from starlette.routing import Router
from starlette.types import Receive, Scope, Send

from connexion.apps.abstract import AbstractApp
from connexion.decorators import StarletteDecorator
from connexion.jsonifier import Jsonifier
from connexion.lifecycle import ConnexionRequest, ConnexionResponse
from connexion.middleware.abstract import RoutedAPI, RoutedMiddleware
from connexion.middleware.lifespan import Lifespan
from connexion.operations import AbstractOperation
from connexion.options import SwaggerUIOptions
from connexion.resolver import Resolver
from connexion.types import MaybeAwaitable
from connexion.uri_parsing import AbstractURIParser

logger = logging.getLogger(__name__)


class AsyncOperation:
    def __init__(
        self,
        fn: t.Callable,
        jsonifier: Jsonifier,
        operation_id: str,
        pythonic_params: bool,
    ) -> None:
        self._fn = fn
        self.jsonifier = jsonifier
        self.operation_id = operation_id
        self.pythonic_params = pythonic_params
        functools.update_wrapper(self, fn)

    @classmethod
    def from_operation(
        cls,
        operation: AbstractOperation,
        *,
        pythonic_params: bool,
        jsonifier: Jsonifier,
    ) -> "AsyncOperation":
        return cls(
            operation.function,
            jsonifier=jsonifier,
            operation_id=operation.operation_id,
            pythonic_params=pythonic_params,
        )

    @property
    def fn(self) -> t.Callable:
        decorator = StarletteDecorator(
            pythonic_params=self.pythonic_params,
            jsonifier=self.jsonifier,
        )
        return decorator(self._fn)

    async def __call__(
        self, scope: Scope, receive: Receive, send: Send
    ) -> StarletteResponse:
        response = await self.fn()
        return await response(scope, receive, send)


class AsyncApi(RoutedAPI[AsyncOperation]):
    def __init__(
        self,
        *args,
        pythonic_params: bool,
        jsonifier: t.Optional[Jsonifier] = None,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.pythonic_params = pythonic_params
        self.jsonifier = jsonifier or Jsonifier()
        self.router = Router()
        self.add_paths()

    def make_operation(self, operation: AbstractOperation) -> AsyncOperation:
        return AsyncOperation.from_operation(
            operation, pythonic_params=self.pythonic_params, jsonifier=self.jsonifier
        )


class AsyncASGIApp(RoutedMiddleware[AsyncApi]):

    api_cls = AsyncApi

    def __init__(self) -> None:
        self.apis: t.Dict[str, t.List[AsyncApi]] = {}
        self.operations: t.Dict[str, AsyncOperation] = {}
        self.router = Router()
        super().__init__(self.router)

    def add_api(self, *args, name: str = None, **kwargs):
        api = super().add_api(*args, **kwargs)

        if name is not None:
            self.router.mount(api.base_path, api.router, name=name)
        else:
            self.router.mount(api.base_path, api.router)
        return api

    def add_url_rule(
        self,
        rule,
        endpoint: str = None,
        view_func: t.Callable = None,
        methods: t.List[str] = None,
        **options,
    ):
        self.router.add_route(rule, endpoint=view_func, name=endpoint, methods=methods)


class AsyncApp(AbstractApp):
    """Connexion Application based on ConnexionMiddleware wrapping a async Connexion application
    based on starlette tools."""

    def __init__(
        self,
        import_name: str,
        *,
        lifespan: t.Optional[Lifespan] = None,
        middlewares: t.Optional[list] = None,
        specification_dir: t.Union[pathlib.Path, str] = "",
        arguments: t.Optional[dict] = None,
        auth_all_paths: t.Optional[bool] = None,
        jsonifier: t.Optional[Jsonifier] = None,
        pythonic_params: t.Optional[bool] = None,
        resolver: t.Optional[t.Union[Resolver, t.Callable]] = None,
        resolver_error: t.Optional[int] = None,
        strict_validation: t.Optional[bool] = None,
        swagger_ui_options: t.Optional[SwaggerUIOptions] = None,
        uri_parser_class: t.Optional[AbstractURIParser] = None,
        validate_responses: t.Optional[bool] = None,
        validator_map: t.Optional[dict] = None,
        security_map: t.Optional[dict] = None,
    ) -> None:
        """
        :param import_name: The name of the package or module that this object belongs to. If you
            are using a single module, __name__ is always the correct value. If you however are
            using a package, it’s usually recommended to hardcode the name of your package there.
        :param lifespan: A lifespan context function, which can be used to perform startup and
            shutdown tasks.
        :param middlewares: The list of middlewares to wrap around the application. Defaults to
            :obj:`middleware.main.ConnexionMiddleware.default_middlewares`
        :param specification_dir: The directory holding the specification(s). The provided path
            should either be absolute or relative to the root path of the application. Defaults to
            the root path.
        :param arguments: Arguments to substitute the specification using Jinja.
        :param auth_all_paths: whether to authenticate not paths not defined in the specification.
            Defaults to False.
        :param jsonifier: Custom jsonifier to overwrite json encoding for json responses.
        :param pythonic_params: When True, CamelCase parameters are converted to snake_case and an
            underscore is appended to any shadowed built-ins. Defaults to False.
        :param resolver: Callable that maps operationId to a function or instance of
            :class:`resolver.Resolver`.
        :param resolver_error: Error code to return for operations for which the operationId could
            not be resolved. If no error code is provided, the application will fail when trying to
            start.
        :param strict_validation: When True, extra form or query parameters not defined in the
            specification result in a validation error. Defaults to False.
        :param swagger_ui_options: Instance of :class:`options.ConnexionOptions` with
            configuration options for the swagger ui.
        :param uri_parser_class: Class to use for uri parsing. See :mod:`uri_parsing`.
        :param validate_responses: Whether to validate responses against the specification. This has
            an impact on performance. Defaults to False.
        :param validator_map: A dictionary of validators to use. Defaults to
            :obj:`validators.VALIDATOR_MAP`.
        :param security_map: A dictionary of security handlers to use. Defaults to
            :obj:`security.SECURITY_HANDLERS`
        """
        self._middleware_app: AsyncASGIApp = AsyncASGIApp()

        super().__init__(
            import_name,
            lifespan=lifespan,
            middlewares=middlewares,
            specification_dir=specification_dir,
            arguments=arguments,
            auth_all_paths=auth_all_paths,
            jsonifier=jsonifier,
            pythonic_params=pythonic_params,
            resolver=resolver,
            resolver_error=resolver_error,
            strict_validation=strict_validation,
            swagger_ui_options=swagger_ui_options,
            uri_parser_class=uri_parser_class,
            validate_responses=validate_responses,
            validator_map=validator_map,
            security_map=security_map,
        )

    def add_url_rule(
        self, rule, endpoint: str = None, view_func: t.Callable = None, **options
    ):
        self._middleware_app.add_url_rule(
            rule, endpoint=endpoint, view_func=view_func, **options
        )

    def add_error_handler(
        self,
        code_or_exception: t.Union[int, t.Type[Exception]],
        function: t.Callable[
            [ConnexionRequest, Exception], MaybeAwaitable[ConnexionResponse]
        ],
    ) -> None:
        self.middleware.add_error_handler(code_or_exception, function)



"""
Module containing all code related to json schema validation.
"""

import contextlib
import io
import os
import typing as t
import urllib.parse
import urllib.request
from collections.abc import Mapping
from copy import deepcopy

import requests
import yaml
from jsonschema import Draft4Validator, RefResolver
from jsonschema.exceptions import RefResolutionError, ValidationError  # noqa
from jsonschema.validators import extend

from .utils import deep_get


class ExtendedSafeLoader(yaml.SafeLoader):
    """Extends the yaml SafeLoader to coerce all keys to string so the result is valid json."""

    def __init__(self, stream):
        self.original_construct_mapping = self.construct_mapping
        self.construct_mapping = self.extended_construct_mapping
        super().__init__(stream)

    def extended_construct_mapping(self, node, deep=False):
        data = self.original_construct_mapping(node, deep)
        return {str(key): data[key] for key in data}


class FileHandler:
    """Handler to resolve file refs."""

    def __call__(self, uri):
        filepath = self._uri_to_path(uri)
        with open(filepath) as fh:
            return yaml.load(fh, ExtendedSafeLoader)

    @staticmethod
    def _uri_to_path(uri):
        parsed = urllib.parse.urlparse(uri)
        host = "{0}{0}{mnt}{0}".format(os.path.sep, mnt=parsed.netloc)
        return os.path.abspath(
            os.path.join(host, urllib.request.url2pathname(parsed.path))
        )


class URLHandler:
    """Handler to resolve url refs."""

    def __call__(self, uri):
        response = requests.get(uri)
        response.raise_for_status()

        data = io.StringIO(response.text)
        with contextlib.closing(data) as fh:
            return yaml.load(fh, ExtendedSafeLoader)


handlers = {
    "http": URLHandler(),
    "https": URLHandler(),
    "file": FileHandler(),
    "": FileHandler(),
}


def resolve_refs(spec, store=None, base_uri=""):
    """
    Resolve JSON references like {"$ref": <some URI>} in a spec.
    Optionally takes a store, which is a mapping from reference URLs to a
    dereferenced objects. Prepopulating the store can avoid network calls.
    """
    spec = deepcopy(spec)
    store = store or {}
    resolver = RefResolver(base_uri, spec, store, handlers=handlers)

    def _do_resolve(node):
        if isinstance(node, Mapping) and "$ref" in node:
            path = node["$ref"][2:].split("/")
            try:
                # resolve known references
                retrieved = deep_get(spec, path)
                node.update(retrieved)
                if isinstance(retrieved, Mapping) and "$ref" in retrieved:
                    node = _do_resolve(node)
                node.pop("$ref", None)
                return node
            except KeyError:
                # resolve external references
                with resolver.resolving(node["$ref"]) as resolved:
                    return _do_resolve(resolved)
        elif isinstance(node, Mapping):
            for k, v in node.items():
                node[k] = _do_resolve(v)
        elif isinstance(node, (list, tuple)):
            for i, _ in enumerate(node):
                node[i] = _do_resolve(node[i])
        return node

    res = _do_resolve(spec)
    return res


def format_error_with_path(exception: ValidationError) -> str:
    """Format a `ValidationError` with path to error."""
    error_path = ".".join(str(item) for item in exception.path)
    error_path_msg = f" - '{error_path}'" if error_path else ""
    return error_path_msg


def allow_nullable(validation_fn: t.Callable) -> t.Callable:
    """Extend an existing validation function, so it allows nullable values to be null."""

    def nullable_validation_fn(validator, to_validate, instance, schema):
        if instance is None and (
            schema.get("x-nullable") is True or schema.get("nullable")
        ):
            return

        yield from validation_fn(validator, to_validate, instance, schema)

    return nullable_validation_fn


def validate_writeOnly(validator, wo, instance, schema):
    yield ValidationError("Property is write-only")


NullableTypeValidator = allow_nullable(Draft4Validator.VALIDATORS["type"])
NullableEnumValidator = allow_nullable(Draft4Validator.VALIDATORS["enum"])

Draft4RequestValidator = extend(
    Draft4Validator,
    {
        "type": NullableTypeValidator,
        "enum": NullableEnumValidator,
    },
)

Draft4ResponseValidator = extend(
    Draft4Validator,
    {
        "type": NullableTypeValidator,
        "enum": NullableEnumValidator,
        "writeOnly": validate_writeOnly,
        "x-writeOnly": validate_writeOnly,
    },
)



import types
import typing as t

# Maybe Awaitable
_ReturnType = t.TypeVar("_ReturnType")
MaybeAwaitable = t.Union[t.Awaitable[_ReturnType], _ReturnType]

# WSGIApp
Environ = t.Mapping[str, object]

_WriteCallable = t.Callable[[bytes], t.Any]
_ExcInfo = t.Tuple[type, BaseException, types.TracebackType]

_StartResponseCallable = t.Callable[
    [
        str,  # status
        t.Sequence[t.Tuple[str, str]],  # response headers
    ],
    _WriteCallable,  # write() callable
]
_StartResponseCallableWithExcInfo = t.Callable[
    [
        str,  # status
        t.Sequence[t.Tuple[str, str]],  # response headers
        t.Optional[_ExcInfo],  # exc_info
    ],
    _WriteCallable,  # write() callable
]
StartResponse = t.Union[_StartResponseCallable, _StartResponseCallableWithExcInfo]
ResponseStream = t.Iterable[bytes]

WSGIApp = t.Callable[[Environ, StartResponse], ResponseStream]



from contextvars import ContextVar

from starlette.types import Receive, Scope
from werkzeug.local import LocalProxy

from connexion.lifecycle import ConnexionRequest
from connexion.operations import AbstractOperation

UNBOUND_MESSAGE = (
    "Working outside of operation context. Make sure your app is wrapped in a "
    "ContextMiddleware and you're processing a request while accessing the context."
)


_context: ContextVar[dict] = ContextVar("CONTEXT")
context = LocalProxy(_context, unbound_message=UNBOUND_MESSAGE)

_operation: ContextVar[AbstractOperation] = ContextVar("OPERATION")
operation = LocalProxy(_operation, unbound_message=UNBOUND_MESSAGE)

_receive: ContextVar[Receive] = ContextVar("RECEIVE")
receive = LocalProxy(_receive, unbound_message=UNBOUND_MESSAGE)

_scope: ContextVar[Scope] = ContextVar("SCOPE")
scope = LocalProxy(_scope, unbound_message=UNBOUND_MESSAGE)

request = LocalProxy(
    lambda: ConnexionRequest(scope, receive), unbound_message=UNBOUND_MESSAGE
)



"""
This module defines a decorator to convert request parameters to arguments for the view function.
"""
import abc
import asyncio
import builtins
import functools
import keyword
import logging
import re
import typing as t
from copy import copy, deepcopy

import inflection

from connexion.context import context, operation
from connexion.frameworks.abstract import Framework
from connexion.http_facts import FORM_CONTENT_TYPES
from connexion.lifecycle import ConnexionRequest, WSGIRequest
from connexion.operations import AbstractOperation, Swagger2Operation
from connexion.utils import (
    deep_merge,
    inspect_function_arguments,
    is_null,
    is_nullable,
    make_type,
)

logger = logging.getLogger(__name__)

CONTEXT_NAME = "context_"


class BaseParameterDecorator:
    def __init__(
        self,
        *,
        framework: t.Type[Framework],
        pythonic_params: bool = False,
    ) -> None:
        self.framework = framework
        self.sanitize_fn = pythonic if pythonic_params else sanitized

    def _maybe_get_body(
        self,
        request: t.Union[WSGIRequest, ConnexionRequest],
        *,
        arguments: t.List[str],
        has_kwargs: bool,
    ) -> t.Any:
        body_name = self.sanitize_fn(operation.body_name(request.content_type))
        # Pass form contents separately for Swagger2 for backward compatibility with
        # Connexion 2 Checking for body_name is not enough
        if (body_name in arguments or has_kwargs) or (
            request.mimetype in FORM_CONTENT_TYPES
            and isinstance(operation, Swagger2Operation)
        ):
            return request.get_body()
        else:
            return None

    @abc.abstractmethod
    def __call__(self, function: t.Callable) -> t.Callable:
        raise NotImplementedError


class SyncParameterDecorator(BaseParameterDecorator):
    def __call__(self, function: t.Callable) -> t.Callable:
        unwrapped_function = unwrap_decorators(function)
        arguments, has_kwargs = inspect_function_arguments(unwrapped_function)

        @functools.wraps(function)
        def wrapper(request: WSGIRequest) -> t.Any:
            request_body = self._maybe_get_body(
                request, arguments=arguments, has_kwargs=has_kwargs
            )

            kwargs = prep_kwargs(
                request,
                request_body=request_body,
                files=request.files(),
                arguments=arguments,
                has_kwargs=has_kwargs,
                sanitize=self.sanitize_fn,
            )

            return function(**kwargs)

        return wrapper


class AsyncParameterDecorator(BaseParameterDecorator):
    def __call__(self, function: t.Callable) -> t.Callable:
        unwrapped_function = unwrap_decorators(function)
        arguments, has_kwargs = inspect_function_arguments(unwrapped_function)

        @functools.wraps(function)
        async def wrapper(request: ConnexionRequest) -> t.Any:
            request_body = self._maybe_get_body(
                request, arguments=arguments, has_kwargs=has_kwargs
            )

            while asyncio.iscoroutine(request_body):
                request_body = await request_body

            kwargs = prep_kwargs(
                request,
                request_body=request_body,
                files=await request.files(),
                arguments=arguments,
                has_kwargs=has_kwargs,
                sanitize=self.sanitize_fn,
            )

            return await function(**kwargs)

        return wrapper


def prep_kwargs(
    request: t.Union[WSGIRequest, ConnexionRequest],
    *,
    request_body: t.Any,
    files: t.Dict[str, t.Any],
    arguments: t.List[str],
    has_kwargs: bool,
    sanitize: t.Callable,
) -> dict:
    kwargs = get_arguments(
        operation,
        path_params=request.path_params,
        query_params=request.query_params,
        body=request_body,
        files=files,
        arguments=arguments,
        has_kwargs=has_kwargs,
        sanitize=sanitize,
        content_type=request.mimetype,
    )

    # optionally convert parameter variable names to un-shadowed, snake_case form
    kwargs = {sanitize(k): v for k, v in kwargs.items()}

    # add context info (e.g. from security decorator)
    for key, value in context.items():
        if has_kwargs or key in arguments:
            kwargs[key] = value
        else:
            logger.debug("Context parameter '%s' not in function arguments", key)
    # attempt to provide the request context to the function
    if CONTEXT_NAME in arguments:
        kwargs[CONTEXT_NAME] = context

    return kwargs


def unwrap_decorators(function: t.Callable) -> t.Callable:
    """Unwrap decorators to return the original function."""
    while hasattr(function, "__wrapped__"):
        function = function.__wrapped__  # type: ignore
    return function


def snake_and_shadow(name: str) -> str:
    """
    Converts the given name into Pythonic form. Firstly it converts CamelCase names to snake_case. Secondly it looks to
    see if the name matches a known built-in and if it does it appends an underscore to the name.
    :param name: The parameter name
    """
    snake = inflection.underscore(name)
    if snake in builtins.__dict__ or keyword.iskeyword(snake):
        return f"{snake}_"
    return snake


def sanitized(name: str) -> str:
    return name and re.sub(
        "^[^a-zA-Z_]+", "", re.sub("[^0-9a-zA-Z_]", "", re.sub(r"\[(?!])", "_", name))
    )


def pythonic(name: str) -> str:
    name = name and snake_and_shadow(name)
    return sanitized(name)


def get_arguments(
    operation: AbstractOperation,
    *,
    path_params: dict,
    query_params: dict,
    body: t.Any,
    files: dict,
    arguments: t.List[str],
    has_kwargs: bool,
    sanitize: t.Callable,
    content_type: str,
) -> t.Dict[str, t.Any]:
    """
    get arguments for handler function
    """
    ret = {}
    ret.update(_get_path_arguments(path_params, operation=operation, sanitize=sanitize))
    ret.update(
        _get_query_arguments(
            query_params,
            operation=operation,
            arguments=arguments,
            has_kwargs=has_kwargs,
            sanitize=sanitize,
        )
    )

    if operation.method.upper() == "TRACE":
        # TRACE requests MUST NOT include a body (RFC7231 section 4.3.8)
        return ret

    ret.update(
        _get_body_argument(
            body,
            operation=operation,
            arguments=arguments,
            has_kwargs=has_kwargs,
            sanitize=sanitize,
            content_type=content_type,
        )
    )
    body_schema = operation.body_schema(content_type)
    ret.update(_get_file_arguments(files, arguments, body_schema, has_kwargs))
    return ret


def _get_path_arguments(
    path_params: dict, *, operation: AbstractOperation, sanitize: t.Callable
) -> dict:
    """
    Extract handler function arguments from path parameters
    """
    kwargs = {}

    path_definitions = {
        parameter["name"]: parameter
        for parameter in operation.parameters
        if parameter["in"] == "path"
    }

    for name, value in path_params.items():
        sanitized_key = sanitize(name)
        if name in path_definitions:
            kwargs[sanitized_key] = _get_val_from_param(value, path_definitions[name])
        else:  # Assume path params mechanism used for injection
            kwargs[sanitized_key] = value
    return kwargs


def _get_val_from_param(value: t.Any, param_definitions: t.Dict[str, dict]) -> t.Any:
    """Cast a value according to its definition in the specification."""
    param_schema = param_definitions.get("schema", param_definitions)

    if is_nullable(param_schema) and is_null(value):
        return None

    if param_schema["type"] == "array":
        type_ = param_schema["items"]["type"]
        format_ = param_schema["items"].get("format")
        return [make_type(part, type_, format_) for part in value]
    else:
        type_ = param_schema["type"]
        format_ = param_schema.get("format")
        return make_type(value, type_, format_)


def _get_query_arguments(
    query_params: dict,
    *,
    operation: AbstractOperation,
    arguments: t.List[str],
    has_kwargs: bool,
    sanitize: t.Callable,
) -> dict:
    """
    extract handler function arguments from the query parameters
    """
    query_definitions = {
        parameter["name"]: parameter
        for parameter in operation.parameters
        if parameter["in"] == "query"
    }

    default_query_params = _get_query_defaults(query_definitions)

    query_arguments = deepcopy(default_query_params)
    query_arguments = deep_merge(query_arguments, query_params)
    return _query_args_helper(
        query_definitions, query_arguments, arguments, has_kwargs, sanitize
    )


def _get_query_defaults(query_definitions: t.Dict[str, dict]) -> t.Dict[str, t.Any]:
    """Get the default values for the query parameter from the parameter definition."""
    defaults = {}
    for k, v in query_definitions.items():
        try:
            if "default" in v:
                defaults[k] = v["default"]
            elif v["schema"]["type"] == "object":
                defaults[k] = _get_default_obj(v["schema"])
            else:
                defaults[k] = v["schema"]["default"]
        except KeyError:
            pass
    return defaults


def _get_default_obj(schema: dict) -> dict:
    try:
        return deepcopy(schema["default"])
    except KeyError:
        properties = schema.get("properties", {})
        return _build_default_obj_recursive(properties, {})


def _build_default_obj_recursive(properties: dict, default_object: dict) -> dict:
    """takes disparate and nested default keys, and builds up a default object"""
    for name, property_ in properties.items():
        if "default" in property_ and name not in default_object:
            default_object[name] = copy(property_["default"])
        elif property_.get("type") == "object" and "properties" in property_:
            default_object.setdefault(name, {})
            default_object[name] = _build_default_obj_recursive(
                property_["properties"], default_object[name]
            )
    return default_object


def _query_args_helper(
    query_definitions: dict,
    query_arguments: dict,
    function_arguments: t.List[str],
    has_kwargs: bool,
    sanitize: t.Callable,
) -> dict:
    result = {}
    for key, value in query_arguments.items():
        sanitized_key = sanitize(key)
        if not has_kwargs and sanitized_key not in function_arguments:
            logger.debug(
                "Query Parameter '%s' (sanitized: '%s') not in function arguments",
                key,
                sanitized_key,
            )
        else:
            logger.debug(
                "Query Parameter '%s' (sanitized: '%s') in function arguments",
                key,
                sanitized_key,
            )
            try:
                query_defn = query_definitions[key]
            except KeyError:  # pragma: no cover
                logger.error(
                    "Function argument '%s' (non-sanitized: %s) not defined in specification",
                    sanitized_key,
                    key,
                )
            else:
                logger.debug("%s is a %s", key, query_defn)
                result.update({sanitized_key: _get_val_from_param(value, query_defn)})
    return result


def _get_body_argument(
    body: t.Any,
    *,
    operation: AbstractOperation,
    arguments: t.List[str],
    has_kwargs: bool,
    sanitize: t.Callable,
    content_type: str,
) -> dict:
    if len(arguments) <= 0 and not has_kwargs:
        return {}

    if not operation.is_request_body_defined:
        return {}

    body_name = sanitize(operation.body_name(content_type))

    if content_type in FORM_CONTENT_TYPES:
        result = _get_body_argument_form(
            body, operation=operation, content_type=content_type
        )

        # Unpack form values for Swagger for compatibility with Connexion 2 behavior
        if content_type in FORM_CONTENT_TYPES and isinstance(
            operation, Swagger2Operation
        ):
            if has_kwargs:
                return result
            else:
                return {
                    sanitize(name): value
                    for name, value in result.items()
                    if sanitize(name) in arguments
                }
    else:
        result = _get_body_argument_json(
            body, operation=operation, content_type=content_type
        )

    if body_name in arguments or has_kwargs:
        return {body_name: result}

    return {}


def _get_body_argument_json(
    body: t.Any, *, operation: AbstractOperation, content_type: str
) -> t.Any:
    # if the body came in null, and the schema says it can be null, we decide
    # to include no value for the body argument, rather than the default body
    if is_nullable(operation.body_schema(content_type)) and is_null(body):
        return None

    if body is None:
        default_body = operation.body_schema(content_type).get("default", {})
        return deepcopy(default_body)

    return body


def _get_body_argument_form(
    body: dict, *, operation: AbstractOperation, content_type: str
) -> dict:
    # now determine the actual value for the body (whether it came in or is default)
    default_body = operation.body_schema(content_type).get("default", {})
    body_props = {
        k: {"schema": v}
        for k, v in operation.body_schema(content_type).get("properties", {}).items()
    }

    # by OpenAPI specification `additionalProperties` defaults to `true`
    # see: https://github.com/OAI/OpenAPI-Specification/blame/3.0.2/versions/3.0.2.md#L2305
    additional_props = operation.body_schema().get("additionalProperties", True)

    body_arg = deepcopy(default_body)
    body_arg.update(body or {})

    if body_props or additional_props:
        return _get_typed_body_values(body_arg, body_props, additional_props)

    return {}


def _get_typed_body_values(body_arg, body_props, additional_props):
    """
    Return a copy of the provided body_arg dictionary
    whose values will have the appropriate types
    as defined in the provided schemas.

    :type body_arg: type dict
    :type body_props: dict
    :type additional_props: dict|bool
    :rtype: dict
    """
    additional_props_defn = (
        {"schema": additional_props} if isinstance(additional_props, dict) else None
    )
    res = {}

    for key, value in body_arg.items():
        try:
            prop_defn = body_props[key]
            res[key] = _get_val_from_param(value, prop_defn)
        except KeyError:  # pragma: no cover
            if not additional_props:
                logger.error(f"Body property '{key}' not defined in body schema")
                continue
            if additional_props_defn is not None:
                value = _get_val_from_param(value, additional_props_defn)
            res[key] = value

    return res


def _get_file_arguments(files, arguments, body_schema: dict, has_kwargs=False):
    results = {}
    for k, v in files.items():
        if not (k in arguments or has_kwargs):
            continue
        if body_schema.get("properties", {}).get(k, {}).get("type") != "array":
            v = v[0]
        results[k] = v

    return results



import contextvars
import typing as t
from unittest.mock import MagicMock

from starlette.types import Receive, Scope

from connexion.context import _context, _operation, _receive, _scope
from connexion.operations import AbstractOperation


class TestContext:
    __test__ = False  # Pytest

    def __init__(
        self,
        *,
        context: dict = None,
        operation: AbstractOperation = None,
        receive: Receive = None,
        scope: Scope = None,
    ) -> None:
        self.context = context if context is not None else self.build_context()
        self.operation = operation if operation is not None else self.build_operation()
        self.receive = receive if receive is not None else self.build_receive()
        self.scope = scope if scope is not None else self.build_scope()

        self.tokens: t.Dict[str, contextvars.Token] = {}

    def __enter__(self) -> None:
        self.tokens["context"] = _context.set(self.context)
        self.tokens["operation"] = _operation.set(self.operation)
        self.tokens["receive"] = _receive.set(self.receive)
        self.tokens["scope"] = _scope.set(self.scope)
        return

    def __exit__(self, type, value, traceback):
        _context.reset(self.tokens["context"])
        _operation.reset(self.tokens["operation"])
        _receive.reset(self.tokens["receive"])
        _scope.reset(self.tokens["scope"])
        return False

    @staticmethod
    def build_context() -> dict:
        return {}

    @staticmethod
    def build_operation() -> AbstractOperation:
        return MagicMock(name="operation")

    @staticmethod
    def build_receive() -> Receive:
        async def receive() -> t.MutableMapping[str, t.Any]:
            return {
                "type": "http.request",
                "body": b"",
            }

        return receive

    @staticmethod
    def build_scope(**kwargs) -> Scope:
        scope = {
            "type": "http",
            "query_string": b"",
            "headers": [(b"Content-Type", b"application/octet-stream")],
        }

        for key, value in kwargs.items():
            scope[key] = value

        return scope



"""
This module defines an AbstractOperation class which implements an abstract Operation interface
and functionality shared between Swagger 2 and OpenAPI 3 specifications.
"""

import abc
import logging

from connexion.utils import all_json

logger = logging.getLogger("connexion.operations.abstract")

DEFAULT_MIMETYPE = "application/json"


class AbstractOperation(metaclass=abc.ABCMeta):

    """
    An API routes requests to an Operation by a (path, method) pair.
    The operation uses a resolver to resolve its handler function.
    We use the provided spec to do a bunch of heavy lifting before
    (and after) we call security_schemes handler.
    The registered handler function ends up looking something like::

        @secure_endpoint
        @validate_inputs
        @deserialize_function_inputs
        @serialize_function_outputs
        @validate_outputs
        def user_provided_handler_function(important, stuff):
            if important:
                serious_business(stuff)
    """

    def __init__(
        self,
        method,
        path,
        operation,
        resolver,
        app_security=None,
        security_schemes=None,
        randomize_endpoint=None,
        uri_parser_class=None,
    ):
        """
        :param method: HTTP method
        :type method: str
        :param path:
        :type path: str
        :param operation: swagger operation object
        :type operation: dict
        :param resolver: Callable that maps operationID to a function
        :param app_security: list of security rules the application uses by default
        :type app_security: list
        :param security_schemes: `Security Definitions Object
            <https://github.com/swagger-api/swagger-spec/blob/master/versions/2.0.md#security-definitions-object>`_
        :type security_schemes: dict
        :param randomize_endpoint: number of random characters to append to operation name
        :type randomize_endpoint: integer
        :param uri_parser_class: class to use for uri parsing
        :type uri_parser_class: AbstractURIParser
        """
        self._method = method
        self._path = path
        self._operation = operation
        self._resolver = resolver
        self._security = operation.get("security", app_security)
        self._security_schemes = security_schemes
        self._uri_parser_class = uri_parser_class
        self._randomize_endpoint = randomize_endpoint
        self._operation_id = self._operation.get("operationId")

        self._resolution = resolver.resolve(self)
        self._operation_id = self._resolution.operation_id

        self._responses = self._operation.get("responses", {})

    @classmethod
    @abc.abstractmethod
    def from_spec(cls, spec, *args, path, method, resolver, **kwargs):
        pass

    @property
    def method(self):
        """
        The HTTP method for this operation (ex. GET, POST)
        """
        return self._method

    @property
    def request_body(self):
        """The request body for this operation"""

    @property
    def is_request_body_defined(self) -> bool:
        """Whether the request body is defined for this operation"""
        return self.request_body != {}

    @property
    def path(self):
        """
        The path of the operation, relative to the API base path
        """
        return self._path

    @property
    def security(self):
        return self._security

    @property
    def security_schemes(self):
        return self._security_schemes

    @property
    def responses(self):
        """
        Returns the responses for this operation
        """
        return self._responses

    @property
    def operation_id(self):
        """
        The operation id used to identify the operation internally to the app
        """
        return self._operation_id

    @property
    def randomize_endpoint(self):
        """
        number of random digits to generate and append to the operation_id.
        """
        return self._randomize_endpoint

    @property
    def router_controller(self):
        """
        The router controller to use (python module where handler functions live)
        """
        return self._router_controller

    @property
    @abc.abstractmethod
    def parameters(self):
        """
        Returns the parameters for this operation
        """

    @property
    @abc.abstractmethod
    def produces(self):
        """
        Content-Types that the operation produces
        """

    @property
    @abc.abstractmethod
    def consumes(self):
        """
        Content-Types that the operation consumes
        """

    @abc.abstractmethod
    def body_name(self, content_type: str) -> str:
        """
        Name of the body in the spec.
        """

    @abc.abstractmethod
    def body_schema(self, content_type: str = None) -> dict:
        """
        The body schema definition for this operation.
        """

    @abc.abstractmethod
    def body_definition(self, content_type: str = None) -> dict:
        """
        The body definition for this operation.
        :rtype: dict
        """

    def response_definition(self, status_code=None, content_type=None):
        """
        response definition for this endpoint
        """
        response_definition = self.responses.get(
            str(status_code), self.responses.get("default", {})
        )
        return response_definition

    @abc.abstractmethod
    def response_schema(self, status_code=None, content_type=None):
        """
        response schema for this endpoint
        """

    @abc.abstractmethod
    def example_response(self, status_code=None, content_type=None):
        """
        Returns an example from the spec
        """

    @abc.abstractmethod
    def get_path_parameter_types(self):
        """
        Returns the types for parameters in the path
        """

    @abc.abstractmethod
    def with_definitions(self, schema):
        """
        Returns the given schema, but with the definitions from the spec
        attached. This allows any remaining references to be resolved by a
        validator (for example).
        """

    def get_mimetype(self):
        """
        If the endpoint has no 'produces' then the default is
        'application/json'.

        :rtype str
        """
        # TODO: don't default
        if all_json(self.produces):
            try:
                return self.produces[0]
            except IndexError:
                return DEFAULT_MIMETYPE
        elif len(self.produces) == 1:
            return self.produces[0]
        else:
            return DEFAULT_MIMETYPE

    @property
    def uri_parser_class(self):
        """
        The uri parser class for this operation.
        """
        return self._uri_parser_class

    @property
    def function(self):
        """
        Resolved function.

        :rtype: types.FunctionType
        """
        return self._resolution.function



"""
This module defines an OpenAPIOperation class, a Connexion operation specific for OpenAPI 3 specs.
"""

import logging

from connexion.datastructures import MediaTypeDict
from connexion.operations.abstract import AbstractOperation
from connexion.uri_parsing import OpenAPIURIParser
from connexion.utils import deep_get

logger = logging.getLogger("connexion.operations.openapi3")


class OpenAPIOperation(AbstractOperation):

    """
    A single API operation on a path.
    """

    def __init__(
        self,
        method,
        path,
        operation,
        resolver,
        path_parameters=None,
        app_security=None,
        security_schemes=None,
        components=None,
        randomize_endpoint=None,
        uri_parser_class=None,
    ):
        """
        This class uses the OperationID identify the module and function that will handle the operation

        From Swagger Specification:

        **OperationID**

        A friendly name for the operation. The id MUST be unique among all operations described in the API.
        Tools and libraries MAY use the operation id to uniquely identify an operation.

        :param method: HTTP method
        :type method: str
        :param path:
        :type path: str
        :param operation: swagger operation object
        :type operation: dict
        :param resolver: Callable that maps operationID to a function
        :param path_parameters: Parameters defined in the path level
        :type path_parameters: list
        :param app_security: list of security rules the application uses by default
        :type app_security: list
        :param security_schemes: `Security Definitions Object
            <https://github.com/swagger-api/swagger-spec/blob/master/versions/2.0.md#security-definitions-object>`_
        :type security_schemes: dict
        :param components: `Components Object
            <https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.1.md#componentsObject>`_
        :type components: dict
        :param randomize_endpoint: number of random characters to append to operation name
        :type randomize_endpoint: integer
        :param uri_parser_class: class to use for uri parsing
        :type uri_parser_class: AbstractURIParser
        """
        self.components = components or {}

        uri_parser_class = uri_parser_class or OpenAPIURIParser

        self._router_controller = operation.get("x-openapi-router-controller")

        super().__init__(
            method=method,
            path=path,
            operation=operation,
            resolver=resolver,
            app_security=app_security,
            security_schemes=security_schemes,
            randomize_endpoint=randomize_endpoint,
            uri_parser_class=uri_parser_class,
        )

        self._parameters = operation.get("parameters", [])
        if path_parameters:
            self._parameters += path_parameters

        self._responses = operation.get("responses", {})

        # TODO figure out how to support multiple mimetypes
        # NOTE we currently just combine all of the possible mimetypes,
        #      but we need to refactor to support mimetypes by response code
        response_content_types = []
        for _, defn in self._responses.items():
            response_content_types += defn.get("content", {}).keys()
        self._produces = response_content_types
        self._consumes = None

        logger.debug("consumes: %s" % self.consumes)
        logger.debug("produces: %s" % self.produces)

    @classmethod
    def from_spec(cls, spec, *args, path, method, resolver, **kwargs):
        return cls(
            method,
            path,
            spec.get_operation(path, method),
            resolver=resolver,
            path_parameters=spec.get_path_params(path),
            app_security=spec.security,
            security_schemes=spec.security_schemes,
            components=spec.components,
            *args,
            **kwargs,
        )

    @property
    def request_body(self):
        return self._operation.get("requestBody", {})

    @property
    def parameters(self):
        return self._parameters

    @property
    def consumes(self):
        if self._consumes is None:
            request_content = self.request_body.get("content", {})
            self._consumes = list(request_content.keys())
        return self._consumes

    @property
    def produces(self):
        return self._produces

    def with_definitions(self, schema: dict):
        if self.components:
            schema.setdefault("schema", {})
            schema["schema"]["components"] = self.components
        return schema

    def response_schema(self, status_code=None, content_type=None):
        response_definition = self.response_definition(status_code, content_type)
        content_definition = response_definition.get("content", response_definition)
        content_definition = content_definition.get(content_type, content_definition)
        if "schema" in content_definition:
            return self.with_definitions(content_definition).get("schema", {})
        return {}

    def example_response(self, status_code=None, content_type=None):
        """
        Returns example response from spec
        """
        # simply use the first/lowest status code, this is probably 200 or 201
        status_code = status_code or sorted(self._responses.keys())[0]

        content_type = content_type or self.get_mimetype()
        examples_path = [str(status_code), "content", content_type, "examples"]
        example_path = [str(status_code), "content", content_type, "example"]
        schema_example_path = [
            str(status_code),
            "content",
            content_type,
            "schema",
            "example",
        ]
        schema_path = [str(status_code), "content", content_type, "schema"]

        try:
            status_code = int(status_code)
        except ValueError:
            status_code = 200
        try:
            # TODO also use example header?
            return (
                list(deep_get(self._responses, examples_path).values())[0]["value"],
                status_code,
            )
        except (KeyError, IndexError):
            pass
        try:
            return (deep_get(self._responses, example_path), status_code)
        except KeyError:
            pass
        try:
            return (deep_get(self._responses, schema_example_path), status_code)
        except KeyError:
            pass

        try:
            return (
                self._nested_example(deep_get(self._responses, schema_path)),
                status_code,
            )
        except KeyError:
            return (None, status_code)

    def _nested_example(self, schema):
        try:
            return schema["example"]
        except KeyError:
            pass
        try:
            # Recurse if schema is an object
            return {
                key: self._nested_example(value)
                for (key, value) in schema["properties"].items()
            }
        except KeyError:
            pass
        try:
            # Recurse if schema is an array
            return [self._nested_example(schema["items"])]
        except KeyError:
            raise

    def get_path_parameter_types(self):
        types = {}
        path_parameters = (p for p in self.parameters if p["in"] == "path")
        for path_defn in path_parameters:
            path_schema = path_defn["schema"]
            if (
                path_schema.get("type") == "string"
                and path_schema.get("format") == "path"
            ):
                # path is special case for type 'string'
                path_type = "path"
            else:
                path_type = path_schema.get("type")
            types[path_defn["name"]] = path_type
        return types

    def body_name(self, _content_type: str) -> str:
        return self.request_body.get("x-body-name", "body")

    def body_schema(self, content_type: str = None) -> dict:
        """
        The body schema definition for this operation.
        """
        return self.body_definition(content_type).get("schema", {})

    def body_definition(self, content_type: str = None) -> dict:
        """
        The body complete definition for this operation.

        **There can be one "body" parameter at most.**
        """
        if self.request_body:
            if content_type is None:
                # TODO: make content type required
                content_type = self.consumes[0]
            if len(self.consumes) > 1:
                logger.warning(
                    "this operation accepts multiple content types, using %s",
                    content_type,
                )
            content_type_dict = MediaTypeDict(self.request_body.get("content", {}))
            res = content_type_dict.get(content_type, {})
            return self.with_definitions(res)
        return {}



"""
This module defines a Connexion specific options class to pass to the Connexion App or API.
"""
import dataclasses
import logging
import typing as t

try:
    from swagger_ui_bundle import swagger_ui_path as default_template_dir
except ImportError:
    default_template_dir = None

NO_UI_MSG = """The swagger_ui directory could not be found.
    Please install connexion with extra install: pip install connexion[swagger-ui]
    or provide the path to your local installation by passing swagger_path=<your path>
"""

logger = logging.getLogger("connexion.options")


@dataclasses.dataclass
class SwaggerUIOptions:
    """Options to configure the Swagger UI.

    :param serve_spec: Whether to serve the Swagger / OpenAPI Specification
    :param spec_path: Where to serve the Swagger / OpenAPI Specification

    :param swagger_ui: Whether to serve the Swagger UI
    :param swagger_ui_path: Where to serve the Swagger UI
    :param swagger_ui_config: Options to configure the Swagger UI. See
          https://swagger.io/docs/open-source-tools/swagger-ui/usage/configuration
          for an overview of the available options.
    :param swagger_ui_template_dir: Directory with static files to use to serve Swagger UI
    :param swagger_ui_template_arguments: Arguments passed to the Swagger UI template. Useful
        when providing your own template dir with additional template arguments.
    """

    serve_spec: bool = True
    spec_path: t.Optional[str] = None

    swagger_ui: bool = True
    swagger_ui_config: dict = dataclasses.field(default_factory=dict)
    swagger_ui_path: str = "/ui"
    swagger_ui_template_dir: t.Optional[str] = None
    swagger_ui_template_arguments: dict = dataclasses.field(default_factory=dict)


class SwaggerUIConfig:
    """Class holding swagger UI specific options."""

    def __init__(
        self,
        options: t.Optional[SwaggerUIOptions] = None,
        oas_version: t.Tuple[int, ...] = (2,),
    ):
        if oas_version >= (3, 0, 0):
            self.spec_path = "/openapi.json"
        else:
            self.spec_path = "/swagger.json"

        if options is not None and not isinstance(options, SwaggerUIOptions):
            raise ValueError(
                f"`swaggger_ui_options` should be of type `SwaggerUIOptions`, "
                f"but received {type(options)} instead."
            )

        self._options = options or SwaggerUIOptions()

    @property
    def openapi_spec_available(self) -> bool:
        """Whether to make the OpenAPI Specification available."""
        return self._options.serve_spec

    @property
    def openapi_spec_path(self) -> str:
        """Path to host the Swagger UI."""
        return self._options.spec_path or self.spec_path

    @property
    def swagger_ui_available(self) -> bool:
        """Whether to make the Swagger UI available."""
        if self._options.swagger_ui and self.swagger_ui_template_dir is None:
            logger.warning(NO_UI_MSG)
            return False
        return self._options.swagger_ui

    @property
    def swagger_ui_path(self) -> str:
        """Path to mount the Swagger UI and make it accessible via a browser."""
        return self._options.swagger_ui_path

    @property
    def swagger_ui_template_dir(self) -> str:
        """Directory with static files to use to serve Swagger UI."""
        return self._options.swagger_ui_template_dir or default_template_dir

    @property
    def swagger_ui_config(self) -> dict:
        """Options to configure the Swagger UI."""
        return self._options.swagger_ui_config

    @property
    def swagger_ui_template_arguments(self) -> dict:
        """Arguments passed to the Swagger UI template."""
        return self._options.swagger_ui_template_arguments
